@@ -4,7 +4,6 @@ import (
 	"context"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"testing"
@@ -68,7 +67,7 @@ func TestReceiveApplicationStream(t *testing.T) {
 		// then
 		require.NoError(t, err)
 		assert.NotEmpty(t, workdir)
-		files, err := ioutil.ReadDir(workdir)
+		files, err := os.ReadDir(workdir)
 		require.NoError(t, err)
 		require.Equal(t, 2, len(files))
 		names := []string{}
